/*
 *			N E T D A E M O N . C
 *
 *  Server MDQS daemon, PDP11 BRLUNIX (based on 4.1a BSD TCP/IP)
 *			DPK at BRL, 1 Jan 82
 *
 *		R E V I S I O N   H I S T O R Y
 */

#include <stdio.h>
#include <signal.h>
#include <pwd.h>
#include "queue.h"
#include <NET/longid.h>
#include <NET/net_types.h>
#include <NET/socket.h>
#include <NET/in.h>

extern	int errno;
extern	char *sys_errlist[];
extern	char *Tmpqdir;
char *MDQSlogin = "mdqs";

struct sockaddr_in addr = { AF_INET };
struct sockaddr_in rmtaddr;	

char	**MDQSserver = 0;	/* Actual smtp server process, full path */
int	Maxconnections = 4;	/* Maximum simultaneous connections */
int	numconnections = 0;	/* Number of currently active connections */
int	debug = 0;		/* If nonzero, give verbose output on stderr */
int	mdqsport = 666;		/* Default MDQS service port */
char	errbuf[BUFSIZ];		/* Logging will be line buffered */
char	programid[40];		/* identification string for log entries */

struct paramtab paramtab[] = {
	"mdqsid",	STRPVAR,	&MDQSlogin,		NULL,
	0,		0,		0,			0
};

main (argc, argv)
int	argc;
char	**argv;
{
	struct	passwd *pw;
	register int	skt;
	int	pid;
	int	i;
	int	alarmtrap();
	char	*them;
	char	hostname[32];
	char	workarea[32];

	rdconf (&paramtab);

	setbuf( stderr, errbuf );
	sprintf( programid, "mdqsdae(%5.5d): ", getpid());
	gethostname( hostname, sizeof(hostname));	/* current hostname */

	if (chdir (Tmpqdir) < 0) {
		log ("Can't chdir to %s (%s)", Tmpqdir, sys_errlist[errno]);
		exit (9);
	}
	if ((pw = getpwnam(MDQSlogin)) == NULL) {
		log ("Can't find MDQSlogin (%s) in /etc/passwd", MDQSlogin);
		exit (8);
	}

	/* Parse args; args will override configuration file... */
	arginit( argc, argv );

	/* Alarms will be used to avoid getting trapped in a wait call */
	signal( SIGALRM, &alarmtrap );

	while (1)
	{
		/*
		 *  Create a socket to accept a MDQS connection on.
		 *  The doloop is due to some sort of IPC bug in 4.1a.
		 */
		if (debug) log("opening socket...");
		i = 60;
		addr.sin_port = htons(mdqsport);   /* swap bytes */
		do{
			skt = socket( SOCK_STREAM, 0, &addr,
				      SO_ACCEPTCONN | SO_KEEPALIVE );
			if( debug ) log( "socket returned %d", skt);
			if( skt < 0 )
				sleep( 5 );
		} while( skt < 0 && i-- > 0 );
		if( skt < 0 ) {
			log( "can't open socket (errno %d)",errno);
			continue;
		}
		if( debug ) log( "socket open on %d", skt );

		/*
		 *  Accept a connection.
		 */
		if( accept( skt, &rmtaddr )) {
			if( debug ) log( "accept error (%d)", errno);
			close( skt );
			sleep(1);
			continue;
		}

		/* We have a valid open connection, start a server... */
		them = raddr( rmtaddr.sin_addr );
		if( them == (-1) )  {
			sprintf( workarea, "%lx", them );
			them = &workarea[0];
		}

		log( "%s started", them);
		if(( pid = fork()) < 0 ) {
			log( "could not fork (%d)", errno);
			close( skt );
		} else if( pid == 0 ) {
			/*
			 *  Child
			 */
			dup2( skt, 0 );
			dup2( skt, 1 );
			if( skt > 1 )
				close( skt );
			setgid (pw->pw_gid);
			setuid (pw->pw_uid);
			execv (*MDQSserver, MDQSserver);
			log( "server exec error (%d)", errno);
			exit (99);
		}

		/*
		 *  Parent
		 */
		close( skt );
		numconnections++;

		/*
		 *  This code collects ZOMBIES and implements load
		 *  limiting by staying in the do loop while the
		 *  Maxconnections active.
		 */
		if( numconnections > 1 || Maxconnections == 1 )
			do {
				alarm( 2 );
				if( wait( &i ) > 0 )
					numconnections--;
				alarm( 0 );
			} while( numconnections >= Maxconnections );
	}
}

alarmtrap() {
	signal( SIGALRM, &alarmtrap );
	return;
}

arginit( argc, argv )
int	argc;
char	**argv;
{
	register int	i;
	register char	*arg;

	if( argc == 1 ) {
		log( "Usage: mdqsdaemon [-d] [-n #connections] [-p port] server [args]");
		exit( 1 );
	}

	for( i = 1; i < argc; i++ ) {
		if( *argv[i] != '-' )
			break;
		arg = argv[i];
		arg++;
		switch( *arg ) {
		case 'd':
			debug++;
			break;

		case 'n':
			if( ++i == argc ) {
				log( "missing number of connections" );
				exit( 99 );
			}
			Maxconnections = atoi( argv[i] );
			if( Maxconnections <= 0 ) {
				log( "Bad number of connections '%s'",
				    argv[i] );
				exit( 99 );
			}
			if (debug) log( "Maxconnection now %d", Maxconnections );
			break;

		case 'p':
			if( ++i == argc ) {
				log( "missing port number" );
				exit( 99 );
			}
			mdqsport = atoi( argv[i] );
			if( mdqsport <= 0 ) {
				log( "Bad port number '%s'", argv[i] );
				exit( 99 );
			}
			if (debug) log( "MDQS port now %d", mdqsport );
			break;
		}
	}

	if( i < argc ) {
		if( access( argv[i], 01 ) < 0 )	{	/* execute privs? */
			log( "Cannot access server program '%s'", argv[i] );
			exit( 99 );
		}
		MDQSserver = &argv[i];
		if( debug ) log( "server is '%s'", *MDQSserver );
	} else {
		log( "MDQSserver program not specified!" );
		exit( 99 );
	}
}

/* VARARGS */
log( fmt, a, b, c, d )
char *fmt;
{
	fputs( programid, stderr );
	fprintf( stderr, fmt, a, b, c, d );
	fputc( '\n', stderr );
	fflush( stderr );
}
